@@ -56,6 +56,8 @@
 
 #define LAZY_PAGES_SOCK_NAME	"lazy-pages.socket"
 
+#define LAZY_PAGES_RESTORE_FINISHED	0x52535446	/* ReSTore Finished */
+
 static mutex_t *lazy_sock_mutex;
 
 struct lazy_iov {
@@ -99,6 +101,8 @@ static LIST_HEAD(lpis);
 static LIST_HEAD(exiting_lpis);
 static LIST_HEAD(pending_lpis);
 static int epollfd;
+static bool restore_finished;
+static struct epoll_rfd lazy_sk_rfd;
 
 static int handle_uffd_event(struct epoll_rfd *lpfd);
 
@@ -114,7 +118,7 @@ static struct lazy_pages_info *lpi_init(void)
 	INIT_LIST_HEAD(&lpi->iovs);
 	INIT_LIST_HEAD(&lpi->reqs);
 	INIT_LIST_HEAD(&lpi->l);
-	lpi->lpfd.revent = handle_uffd_event;
+	lpi->lpfd.read_event = handle_uffd_event;
 
 	return lpi;
 }
@@ -460,6 +464,8 @@ static int drop_iovs(struct lazy_pages_info *lpi, unsigned long addr, int len)
 			addr = start;
 		}
 
+		iov->queued = false;
+
 		/*
 		 * The range completely fits into the current IOV.
 		 * If addr equals iov_base we just "drop" the
@@ -802,7 +808,7 @@ static int uffd_zero(struct lazy_pages_info *lpi, __u64 address, int nr_pages)
 /*
  * Seek for the requested address in the pagemap. If it is found, the
  * subsequent call to pr->page_read will bring us the data. If the
- * address is not found in the pagemap, but no error occured, the
+ * address is not found in the pagemap, but no error occurred, the
  * address should be mapped to zero pfn.
  *
  * Returns 0 for zero pages, 1 for "real" pages and negative value on
@@ -1103,6 +1109,7 @@ static void lazy_pages_summary(struct lazy_pages_info *lpi)
 static int handle_requests(int epollfd, struct epoll_event *events, int nr_fds)
 {
 	struct lazy_pages_info *lpi, *n;
+	/* FIXME -- timeout should decrease over time...  */
 	int poll_timeout = POLL_TIMEOUT;
 	int ret;
 
@@ -1118,6 +1125,10 @@ static int handle_requests(int epollfd, struct epoll_event *events, int nr_fds)
 			continue;
 		}
 
+		/* don't start backround fetch before restore is finished */
+		if (!restore_finished)
+			continue;
+
 		if (poll_timeout)
 			pr_debug("Start handling remaining pages\n");
 
@@ -1148,6 +1159,29 @@ out:
 
 }
 
+int lazy_pages_finish_restore(void)
+{
+	uint32_t fin = LAZY_PAGES_RESTORE_FINISHED;
+	int fd, ret;
+
+	if (!opts.lazy_pages)
+		return 0;
+
+	fd = get_service_fd(LAZY_PAGES_SK_OFF);
+	if (fd < 0) {
+		pr_err("No lazy-pages socket\n");
+		return -1;
+	}
+
+	ret = send(fd, &fin, sizeof(fin), 0);
+	if (ret != sizeof(fin))
+		pr_perror("Failed sending restore finished indication");
+
+	close(fd);
+
+	return ret < 0 ? ret : 0;
+}
+
 static int prepare_lazy_socket(void)
 {
 	int listen;
@@ -1165,6 +1199,44 @@ static int prepare_lazy_socket(void)
 	return listen;
 }
 
+static int lazy_sk_read_event(struct epoll_rfd *rfd)
+{
+	uint32_t fin;
+	int ret;
+
+	ret = recv(rfd->fd, &fin, sizeof(fin), 0);
+	/*
+	 * epoll sets POLLIN | POLLHUP for the EOF case, so we get short
+	 * read just befor hangup_event
+	 */
+	if (!ret)
+		return 0;
+
+	if (ret != sizeof(fin)) {
+		pr_perror("Failed getting restore finished indication");
+		return -1;
+	}
+
+	if (fin != LAZY_PAGES_RESTORE_FINISHED) {
+		pr_err("Unexpected response: %x\n", fin);
+		return -1;
+	}
+
+	restore_finished = true;
+
+	return 0;
+}
+
+static int lazy_sk_hangup_event(struct epoll_rfd *rfd)
+{
+	if (!restore_finished) {
+		pr_err("Restorer unexpectedly closed the connection\n");
+		return -1;
+	}
+
+	return 0;
+}
+
 static int prepare_uffds(int listen, int epollfd)
 {
 	int i;
@@ -1190,7 +1262,12 @@ static int prepare_uffds(int listen, int epollfd)
 			goto close_uffd;
 	}
 
-	close_safe(&client);
+	lazy_sk_rfd.fd = client;
+	lazy_sk_rfd.read_event = lazy_sk_read_event;
+	lazy_sk_rfd.hangup_event = lazy_sk_hangup_event;
+	if (epoll_add_rfd(epollfd, &lazy_sk_rfd))
+		goto close_uffd;
+
 	close(listen);
 	return 0;
 
@@ -1240,7 +1317,12 @@ int cr_lazy_pages(bool daemon)
 	if (close_status_fd())
 		return -1;
 
-	nr_fds = task_entries->nr_tasks + (opts.use_page_server ? 1 : 0);
+	/*
+	 * we poll nr_tasks userfault fds, UNIX socket between lazy-pages
+	 * daemon and the cr-restore, and, optionally TCP socket for
+	 * remote pages
+	 */
+	nr_fds = task_entries->nr_tasks + (opts.use_page_server ? 2 : 1);
 	epollfd = epoll_prepare(nr_fds, &events);
 	if (epollfd < 0)
 		return -1;
